from typing import Any

import numpy as np

from qxmt.datasets.generate.linear import (
    generate_linear_regression_data,
    generate_linear_separable_data,
)


class GeneratedDataLoader:
    """
    This class loads the generated data and label.
    The loaded data and label are returned as numpy arrays (X and y).
    The dataset is generated by the specified task and method.

    Supported task types:
    - classification
    - regression

    Supported generate methods:
    - linear

    Examples:
        >>> loader = GeneratedDataLoader(task_type="classification", generate_method="linear", random_seed=42)
        >>> X, y = loader.load()
        >>> loader = GeneratedDataLoader(task_type="regression", generate_method="linear", random_seed=42)
        >>> X, y = loader.load()
    """

    def __init__(
        self,
        task_type: str,
        generate_method: str,
        random_seed: int,
        params: dict[str, Any] = {},
    ) -> None:
        """Initialize the GeneratedDataLoader.

        Args:
            task_type (str): dataset task type
            generate_method (str): generate dataset method
            random_seed (int): random seed for reproducibility
            params (dict[str, Any]]): additional parameters for dataset generation
        """
        self.task_type = task_type
        self.generate_method = generate_method
        self.random_seed = random_seed
        self.params = params

    def load(self) -> tuple[np.ndarray, np.ndarray]:
        """Load dummy dataset. The dataset is generated by the specified task and method.

        Args:
            task_type (str): dataset task type. "classification" or "regression" is supported.
            generate_method (str): generate dataset method. "linear" is supported.
            random_seed (int): random seed for reproducibility.
            params (dict[str, Any]]): additional parameters for dataset generation.

        Raises:
            ValueError: not supported task_type or generate_method

        Returns:
            tuple[np.ndarray, np.ndarray]: generated data and label
        """
        match (self.task_type, self.generate_method):
            case ("classification", "linear"):
                X, y = generate_linear_separable_data(
                    n_samples=self.params.get("n_samples", 100),
                    n_features=self.params.get("n_features", 2),
                    n_classes=self.params.get("n_classes", 2),
                    noise=self.params.get("noise", 0.1),
                    scale=self.params.get("scale", 1.0),
                    random_seed=self.random_seed,
                )
            case ("regression", "linear"):
                X, y = generate_linear_regression_data(
                    n_samples=self.params.get("n_samples", 100),
                    n_features=self.params.get("n_features", 2),
                    noise=self.params.get("noise", 0.1),
                    scale=self.params.get("scale", 1.0),
                    random_seed=self.random_seed,
                )
            case _:
                raise ValueError(
                    f'Unsupported task_type="{self.task_type}" or generate_method="{self.generate_method}"'
                )

        return X, y



from pathlib import Path
from typing import Any, Optional

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from qxmt.constants import DEFAULT_COLOR_MAP
from qxmt.decorators import notify_long_running
from qxmt.models.qsvc import QSVC


def plot_2d_predicted_result(
    X: np.ndarray,
    y_true: np.ndarray,
    y_pred: np.ndarray,
    feature_cols: Optional[list[str]] = None,
    axis: list[int] = [0, 1],
    save_path: Optional[str | Path] = None,
    **kwargs: Any,
) -> None:
    """Plot predicted result on 2D plane.

    Args:
        X (np.ndarray): feature values.
        y_true (np.ndarray): ground truth labels.
        y_pred (np.ndarray): predicted labels.
        feature_cols (Optional[list[str]], optional): feature column names. Defaults to None.
        axis (list[int], optional): axis to plot (target feature col index). Defaults to [0, 1].
        save_path (Optional[str], optional): save path of graph. Defaults to None.
        **kwargs (Any): additional arguments for plot.
            truth_label (str, optional): label of ground truth. Defaults to "Ground Truth".
            pred_label (str, optional): label of predicted. Defaults to "Predicted".
            title (str, optional): title of the plot. Defaults to '"Groud Truth" VS "Predicted"'.
            colors (Optional[dict[int, str]], optional): color of each class. Defaults to None.
            class_labels (Optional[dict[int, str]], optional): label of each class. Defaults to None.
    """
    assert X.shape[0] == len(y_true) == len(y_pred), "Length of X , y_true and y_pred must be the same."

    if feature_cols is None:
        feature_cols = [f"feature_{i}" for i in range(max(axis) + 1)]

    # set each class color and label
    colors = kwargs.get("colors", sns.color_palette("viridis", n_colors=len(np.unique(y_true))))
    class_labels = kwargs.get("class_labels", {class_value: str(class_value) for class_value in np.unique(y_true)})

    # plot grpoed truth and predicted result
    plt.figure(figsize=(7, 5), tight_layout=True)
    color_labels = []
    for i, class_value in enumerate(np.unique(y_true)):
        groud_subset = X[np.where(y_true == class_value)]
        plt.scatter(
            groud_subset[:, axis[0]],
            groud_subset[:, axis[1]],
            edgecolor=colors[i],
            facecolor="none",
            s=100,
        )

        predicted_subset = X[np.where(y_pred == class_value)]
        plt.scatter(
            predicted_subset[:, axis[0]],
            predicted_subset[:, axis[1]],
            marker="x",
            color=colors[i],
        )

        # legend for colors, it is empty because we want to show only labels
        color_label = plt.scatter([], [], marker="o", color=colors[i], label=f"{class_labels.get(class_value)}")
        color_labels.append(color_label)

    # legend for markers, it is empty because we want to show only labels
    truth_label = plt.scatter(
        [], [], label=str(kwargs.get("truth_label", "Ground Truth")), marker="o", facecolor="none", color="black"
    )
    predict_label = plt.scatter([], [], label=str(kwargs.get("pred_label", "Predicted")), marker="x", color="black")
    marker_legend = plt.legend(
        title="marker type", handles=[truth_label, predict_label], loc="upper right", bbox_to_anchor=(1.4, 1)
    )
    _ = plt.legend(title="Class", handles=color_labels, loc="upper right", bbox_to_anchor=(1.4, 0.8))

    plt.gca().add_artist(marker_legend)

    plt.xlabel(f"{feature_cols[axis[0]]}")
    plt.ylabel(f"{feature_cols[axis[1]]}")

    title = str(kwargs.get("title", '"Groud Truth" VS "Predicted"'))
    plt.title(title)

    if save_path is not None:
        plt.savefig(save_path, bbox_inches="tight", pad_inches=0.1)

    plt.show()


@notify_long_running
def plot_2d_decision_boundaries(
    model: QSVC,
    X: np.ndarray,
    y: np.ndarray,
    grid_resolution: int = 100,
    support_vectors: bool = True,
    feature_cols: Optional[list[str]] = None,
    save_path: Optional[str | Path] = None,
    **kwargs: Any,
) -> None:
    """Plot decision boundaries of QSVC on 2D plane.

    Args:
        model (QSVC): QSVC model.
        X (np.ndarray): Feature values of the data.
        y (np.ndarray): Labels of the data.
        grid_resolution (int, optional): Resolution of the grid. Defaults to 100.
        support_vectors (bool, optional): Plot support vectors or not. Defaults to True.
        feature_cols (Optional[list[str]], optional): Feature column names. Defaults to None.
        save_path (Optional[str | Path], optional): Save path of the graph. Defaults to None.
        **kwargs (Any): Additional arguments for the plot.
            title (str, optional): Title of the plot. Defaults to "Decision boundaries of QSVC".
            cmap (str, optional): Color map for the plot. Defaults to DEFAULT_COLOR_MAP.
    """
    assert X.shape[0] == len(y), "Length of X and y must be the same."

    if isinstance(model, QSVC):
        is_multi_class = len(model.classes_) > 2
    else:
        is_multi_class = len(np.unique(y)) > 2

    if X.shape[1] != 2:
        raise ValueError("This function only supports 2D datasets.")

    if feature_cols is None:
        feature_cols = ["feature_0", "feature_1"]

    _, ax = plt.subplots(figsize=(6, 5), tight_layout=True)
    x_min, x_max = X[:, 0].min(), X[:, 0].max()
    y_min, y_max = X[:, 1].min(), X[:, 1].max()
    margin = 1
    ax.set(xlim=(x_min - margin, x_max + margin), ylim=(y_min - margin, y_max + margin))

    # Generate grid
    xx, yy = np.meshgrid(
        np.linspace(x_min - margin, x_max + margin, grid_resolution),
        np.linspace(y_min - margin, y_max + margin, grid_resolution),
    )
    grid_points = np.c_[xx.ravel(), yy.ravel()]  # Shape (n_grid_points, 2)

    # Compute predictions over the grid
    Z = model.predict(grid_points)
    if not is_multi_class:
        decision_function = model.decision_function(grid_points)

    Z = Z.reshape(xx.shape)

    # Plot decision boundaries
    cmap = kwargs.get("cmap", DEFAULT_COLOR_MAP)
    ax.pcolormesh(xx, yy, Z, cmap=cmap, shading="auto", alpha=0.3, vmin=y.min(), vmax=y.max())

    if not is_multi_class:
        # Plot decision boundary and margins
        Z_decision = decision_function.reshape(xx.shape)
        ax.contour(
            xx,
            yy,
            Z_decision,
            levels=[-1, 0, 1],
            colors=["k", "k", "k"],
            linestyles=["--", "-", "--"],
        )

    if support_vectors and hasattr(model, "support_"):
        # Plot support vectors
        ax.scatter(
            X[model.support_, 0],
            X[model.support_, 1],
            s=150,
            facecolors="none",
            edgecolors="k",
        )

    # Plot samples by color and add legend
    scatter = ax.scatter(X[:, 0], X[:, 1], c=y, s=30, edgecolors="k", cmap=cmap, vmin=y.min(), vmax=y.max())
    plt.xlabel(f"{feature_cols[0]}")
    plt.ylabel(f"{feature_cols[1]}")
    ax.legend(*scatter.legend_elements(), loc="upper right", title="Class", bbox_to_anchor=(1.2, 1))

    title = str(kwargs.get("title", "Decision boundaries of QSVC"))
    ax.set_title(title)

    if save_path is not None:
        plt.savefig(save_path, bbox_inches="tight", pad_inches=0.1)

    plt.show()



import copy
import importlib
import sys
import types
from functools import wraps
from pathlib import Path
from typing import Any, Callable

import yaml

from qxmt import ExperimentConfig


def load_yaml_config(file_path: str | Path) -> dict:
    """Load yaml from configuration file.

    Args:
        file_path (str | Path): path to the configuration file

    Returns:
        dict: configuration
    """
    if Path(file_path).suffix != ".yaml":
        raise ValueError("The configuration file must be a yaml file.")

    if not Path(file_path).exists():
        raise FileNotFoundError(f"File '{file_path}' not found.")

    with open(file_path, "r") as file:
        config = yaml.safe_load(file)

    return config


def load_object_from_yaml(config: dict, dynamic_params: dict = {}, use_cache: bool = False) -> Any:
    """Extract object from yaml configuration and convert it to a function or class instance.

    Args:
        config (dict): configuration of the object
        dynamic_params (dict, optional): dynamic parameters to be passed to the object. Defaults to {}.
        use_cache (bool, optional): use cache when loading the module. Defaults to False.

    Raises:
        ModuleNotFoundError: module not found in the path
        AttributeError: object not found in the module
        TypeError: object is not a class or function

    Returns:
        Any: function or instance of the extracted object
    """
    module_name = config.get("module_name", None)
    object_name = config.get("implement_name", None)
    params = config.get("params", None)
    params = {} if params is None else params
    params.update(dynamic_params)

    try:
        if (not use_cache) and (module_name in sys.modules):
            # load module not use cache
            module = sys.modules[module_name]
            importlib.reload(module)
        else:
            module = importlib.import_module(module_name)
        obj = getattr(module, object_name)
    except ModuleNotFoundError:
        raise ImportError(f"Module '{module_name}' not found.")
    except AttributeError:
        raise AttributeError(f"Object '{object_name}' not found in module '{module_name}'.")

    if isinstance(obj, types.FunctionType):
        # Extracted object is a function
        @wraps(obj)
        def callable_func(*args: Any, **kwargs: Any) -> Callable:
            """Wrapper function that applies the params to the extracted function."""
            combined_params = {**params, **kwargs}  # Merge YAML params with any additional kwargs
            return obj(*args, **combined_params)

        # Set type hints for the wrapper function
        if hasattr(obj, "__annotations__"):
            callable_func.__annotations__ = obj.__annotations__

        return callable_func
    elif isinstance(obj, type):
        # Extracted object is a class instance
        return obj(**params)
    else:
        raise TypeError(f"'{object_name}' is not a class or function.")


def save_experiment_config_to_yaml(
    config: ExperimentConfig,
    save_path: str | Path,
    delete_source_path: bool = True,
) -> None:
    """Save the experiment configuration to a yaml file.

    Args:
        config (ExperimentConfig): experiment configuration
        save_path (str | Path): path to save the configuration file
        delete_source_path (bool, optional): delete the source path in the configuration. Defaults to True.
    """
    save_config = copy.deepcopy(config)

    if save_config.dataset.openml is not None:
        save_config.dataset.openml.save_path = str(save_config.dataset.openml.save_path)

    if save_config.dataset.file is not None:
        save_config.dataset.file.data_path = str(save_config.dataset.file.data_path)
        save_config.dataset.file.label_path = str(save_config.dataset.file.label_path)

    config_dict = save_config.model_dump()
    if delete_source_path:
        del config_dict["path"]

    with open(save_path, "w") as file:
        yaml.dump(config_dict, file, sort_keys=False)



import numpy as np
import pennylane as qml

from qxmt.constants import PENNYLANE_PLATFORM
from qxmt.feature_maps import BaseFeatureMap


class YZCXFeatureMap(BaseFeatureMap):
    """YZCX feature map class.
    Reference: https://arxiv.org/abs/2108.01039

    Args:
        BaseFeatureMap (_type_): base feature map class

    Example:
        >>> import numpy as np
        >>> from qxmt.feature_maps.pennylane.yzcx import YZCXFeatureMap
        >>> feature_map = YZCXFeatureMap(2, 2, 0.1, 0)
        >>> feature_map.draw(x_dim=2)
        0: ──RY(0.03)──RY(3.42)──RZ(0.05)──RZ(5.88)─╭●──RY(0.03)──RY(5.39)──RZ(0.05)──RZ(0.21)
        1: ─────────────────────────────────────────╰X──RY(0.03)──RY(5.13)──RZ(0.05)──RZ(0.02)
        ──────────────────────────────────────────┤
        ──RY(0.03)───RY(4.58)──RZ(0.05)──RZ(1.10)─┤
    """

    def __init__(self, n_qubits: int, reps: int, c: float, seed: int) -> None:
        """ "Initialize the YZCX feature map class.

        Args:
            n_qubits (int): number of qubits
            reps (int): number of repetitions
            c (float): scaling factor
            seed (int): random seed
        """
        super().__init__(PENNYLANE_PLATFORM, n_qubits)
        self.n_qubits: int = n_qubits
        self.reps: int = reps
        self.c: float = c
        self.seed: int = seed

    def feature_map(self, x: np.ndarray) -> None:
        """Create quantum circuit of YZCX feature map.

        Args:
            x (np.ndarray): input data
        """
        rng = np.random.default_rng(self.seed)
        data_idx = 0
        for r_idx in range(self.reps):
            for i in range(self.n_qubits):
                # Apply rotaion Y gate by data value and random angle
                qml.RY(self.c * x[data_idx % len(x)], wires=i)
                ry_angle = 2.0 * np.pi * rng.random()
                qml.RY(ry_angle, wires=i)
                data_idx += 1

                # Apply rotaion Z gate by data value and random angle
                qml.RZ(self.c * x[data_idx % len(x)], wires=i)
                rz_angle = 2.0 * np.pi * rng.random()
                qml.RZ(rz_angle, wires=i)
                data_idx += 1

                # Apply CNOT gate based on the current repetition and qubit index
                if (i % 2 == r_idx % 2) and (i + 1 < self.n_qubits):
                    qml.CNOT(wires=[i, i + 1])



from textwrap import dedent

diff_desc_system_prompt = dedent(
    """
    # Role:
    You are a code analysis assistant specialized in summarizing GitHub commit changes.
    Your task is to analyze the given code edits and provide a concise summary of the modifications.
    Focus on describing what was changed, added, removed, or refactored, including the purpose and impact of these
    changes. Provide the summary in clear, professional English, highlighting key changes that are relevant to
    developers reviewing the commit.

    # Task:
    Please generate one or two sentences as a description for the input diff code.
    The input diff code contains two parts: ADD CODE and REMOVE CODE.
    The description should include a clear and concise summary of the code changes.
    """
)


diff_desc_user_prompt = dedent(
    """
    code diff:

    --- ADD CODE ---
    {add_code}

    --- REMOVE CODE ---
    {remove_code}
    """
)



import multiprocessing as mp
from abc import ABC, abstractmethod
from pathlib import Path
from types import FunctionType
from typing import Callable, Optional, cast

import h5py
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable
from rich.progress import Progress

from qxmt.constants import DEFAULT_N_JOBS
from qxmt.devices.base import BaseDevice
from qxmt.exceptions import DeviceSettingError
from qxmt.feature_maps.base import BaseFeatureMap, FeatureMapFromFunc
from qxmt.kernels.sampling import generate_all_observable_states


class BaseKernel(ABC):
    """
    Base kernel class for quantum kernel computation.
    This class is used to compute the kernel value between two samples.
    If defining a custom kernel, inherit this class and implement the `compute()` method.
    The feature map used to compute the kernel value is defined in the constructor.
    It is possible to use a feature map instance or a function that defines the feature map circuit.

    Examples:
        >>> import numpy as np
        >>> from typing import Callable
        >>> from qxmt.kernels.base import BaseKernel
        >>> from qxmt.feature_maps.pennylane.defaults import ZZFeatureMap
        >>> from qxmt.configs import DeviceConfig
        >>> from qxmt.devices.base import BaseDevice
        >>> from qxmt.devices.builder import DeviceBuilder
        >>> config = DeviceConfig(
        ...     platform="pennylane",
        ...     name="default.qubit",
        ...     n_qubits=2,
        ...     shots=1000,
        >>> )
        >>> device = DeviceBuilder(config).build()
        >>> feature_map = ZZFeatureMap(2, 2)
        >>> class CustomKernel(BaseKernel):
        ...     def __init__(self, device: BaseDevice, feature_map: Callable[[np.ndarray], None]) -> None:
        ...         super().__init__(device, feature_map)
        ...
        ...     def compute(self, x1: np.ndarray, x2: np.ndarray) -> float:
        ...         return np.dot(x1, x2)
        >>> kernel = CustomKernel(device, feature_map)
        >>> x1 = np.random.rand(2)
        >>> x2 = np.random.rand(2)
        >>> kernel.compute(x1, x2)
        0.28
    """

    def __init__(
        self,
        device: BaseDevice,
        feature_map: BaseFeatureMap | Callable[[np.ndarray], None],
    ) -> None:
        """Initialize the kernel class.

        Args:
            device (BaseDevice): device instance for quantum computation
            feature_map (BaseFeatureMap | Callable[[np.ndarray], None]): feature map instance or function
        """
        self.device: BaseDevice = device
        self.platform: str = self.device.platform
        self.n_qubits: int = self.device.n_qubits
        self.feature_map: BaseFeatureMap = self._set_feature_map(feature_map)
        self.is_sampling: bool = (self.device.shots is not None) and (self.device.shots > 0)

    def _set_feature_map(self, feature_map: BaseFeatureMap | Callable[[np.ndarray], None]) -> BaseFeatureMap:
        """Set the feature map instance of the BaseFeatureMap.

        Args:
            feature_map (BaseFeatureMap | Callable[[np.ndarray], None]): feature map instance or function

        Returns:
            BaseFeatureMap: feature map instance
        """
        if isinstance(feature_map, FunctionType):
            return FeatureMapFromFunc(self.platform, self.n_qubits, feature_map)
        else:
            return cast(BaseFeatureMap, feature_map)

    @abstractmethod
    def compute(self, x1: np.ndarray, x2: np.ndarray) -> tuple[float, np.ndarray]:
        """Compute kernel value between two samples.

        Args:
            x1 (np.ndarray): array of one sample
            x2 (np.ndarray): array of one sample

        Returns:
            tuple[float, np.ndarray]: kernel value and probability distribution
        """
        pass

    def _compute_entry(
        self, i: int, j: int, x_array_1: np.ndarray, x_array_2: np.ndarray, progress_queue: mp.Queue
    ) -> tuple[int, int, tuple[float, np.ndarray] | Exception]:
        """Compute each entry of the kernel matrix.
        This method is used for parallel computation of the kernel matrix.
        If an error occurs in the self.compute() method, it is handled and returned as an exception.

        Args:
            i (int): row index of kernel matrix
            j (int): column index of kernel matrix
            x_array_1 (np.ndarray): input array 1 for kernel computation
            x_array_2 (np.ndarray): input array 2 for kernel computation
            progress_queue (mp.Queue): queue for tracking the progress

        Returns:
            tuple[int, int, tuple[float, np.ndarray] | Exception]:
                row index, column index, kernel value and probability distribution

        Raises:
            Exception: error in the self.compute() method
        """
        try:
            result = self.compute(x_array_1[i], x_array_2[j])
            progress_queue.put(1)
            return i, j, result
        except Exception as e:
            progress_queue.put(1)
            return i, j, e

    def compute_matrix(
        self,
        x_array_1: np.ndarray,
        x_array_2: np.ndarray,
        return_shots_resutls: bool = False,
        n_jobs: int = DEFAULT_N_JOBS,
        bar_label: str = "",
    ) -> tuple[np.ndarray, Optional[np.ndarray]]:
        """Default implementation of kernel matrix computation.
        Due to the parallel computation, raise an error is delayed until the end of the computation.

        Args:
            x_array_1 (np.ndarray): array of samples (ex: training data)
            x_array_2 (np.ndarray): array of samples (ex: test data)
            return_shots_resutls (bool, optional): return the shot results. Defaults to False.
            n_jobs (int, optional): number of jobs for parallel computation. Defaults to DEFAULT_N_JOBS.
            bar_label (str, optional): label for the progress bar. Defaults to "".

        Returns:
            np.ndarray: computed kernel matrix

        Raises:
            Exception: error in the self.compute() method
        """

        # compute each entry of the kernel matrix in parallel
        n_samples_1 = len(x_array_1)
        n_samples_2 = len(x_array_2)

        # parallel computation for each entry of the kernel matrix
        tasks = [(i, j, x_array_1, x_array_2) for i in range(len(x_array_1)) for j in range(len(x_array_2))]
        with mp.Manager() as manager:
            progress_queue = manager.Queue()
            with Progress() as progress:
                bar_label = f" ({bar_label})" if bar_label else ""
                task_progress = progress.add_task(f"Computing Kernel Matrix{bar_label}", total=len(tasks))

                with mp.Pool(processes=n_jobs) as pool:
                    results = pool.starmap_async(
                        self._compute_entry,
                        [(i, j, x_array_1, x_array_2, progress_queue) for (i, j, x_array_1, x_array_2) in tasks],
                    )

                    # track progress
                    completed = 0
                    while not progress.finished:
                        progress_queue.get()
                        completed += 1
                        progress.update(task_progress, completed=completed)

                    # get all process results
                    results.wait()
                    final_results = results.get()

                    # finalize progress bar
                    progress.update(task_progress, completed=len(tasks))
                    progress.refresh()

        # initialize the shots results matrix when return_shots_resutls is True and sampling is enabled
        if self.is_sampling and return_shots_resutls:
            num_state = 2**self.device.n_qubits
            shots_matrix = np.zeros((n_samples_1, n_samples_2, num_state))
        else:
            shots_matrix = None

        kernel_matrix = np.zeros((n_samples_1, n_samples_2))
        for i, j, result in final_results:
            if isinstance(result, Exception):
                # raise error in self.compute() method
                raise result
            else:
                # success to compute the kernel value
                kernel_matrix[i, j] = result[0]

                if shots_matrix is not None:
                    shots_matrix[i, j] = result[1]

        return kernel_matrix, shots_matrix

    def save_shots_results(self, probs_matrix: np.ndarray, save_path: str | Path) -> None:
        """Save the shot results to a file.
        probs_matrix contains the probability distribution of the observable states for each sample.
        expected shape of probs_matrix: (n_samples_1, n_samples_2, num_state)

        Args:
            probs_matrix (np.ndarray): probability distribution of the observable states
            save_path (str | Path): save path for the shot results

        Raises:
            DeviceSettingError: not sampling mode
            ValueError: save path extension not ".h5"
            ValueError: state labels and probs length mismatch
        """
        if not self.is_sampling:
            raise DeviceSettingError("Shots must be set to a positive integer value to save the shot results.")

        if Path(save_path).suffix != ".h5":
            raise ValueError("The save path must be a .h5 file.")

        state_labels = generate_all_observable_states(self.n_qubits, state_pattern="01")
        if probs_matrix.shape[2] != len(state_labels):
            raise ValueError(
                "The length of the state of probability distribution must be equal to the number of observable states."
            )

        with h5py.File(save_path, "w") as f:
            probs_dataset = f.create_dataset("probs", data=probs_matrix, compression="gzip", dtype="float32")

            metadata_group = f.create_group("metadata")
            metadata_group.attrs["platform"] = self.platform
            metadata_group.attrs["n_qubits"] = self.n_qubits
            metadata_group.attrs["shots"] = self.device.shots
            metadata_group.attrs["state_labels"] = state_labels
            metadata_group.attrs["n_elements"] = probs_dataset.shape[0] * probs_dataset.shape[1]

    def plot_matrix(
        self,
        x_array_1: np.ndarray,
        x_array_2: np.ndarray,
        save_path: Optional[str | Path] = None,
        n_jobs: int = DEFAULT_N_JOBS,
    ) -> None:
        """Plot kernel matrix for given samples.
        Caluculation of kernel values is performed in parallel.

        Args:
            x_array_1 (np.ndarray): array of samples (ex: training data)
            x_array_2 (np.ndarray): array of samples (ex: test data)
            save_path (Optional[str | Path], optional): save path for the plot. Defaults to None.
            n_jobs (int, optional): number of jobs for parallel computation. Defaults to DEFAULT_N_JOBS.
        """

        kernel_matrix, _ = self.compute_matrix(x_array_1, x_array_2, return_shots_resutls=False, n_jobs=n_jobs)
        plt.imshow(np.asmatrix(kernel_matrix), interpolation="nearest", origin="upper", cmap="viridis")
        plt.colorbar()
        plt.title("Kernel matrix")
        if save_path is not None:
            plt.savefig(save_path)

        plt.show()

    def plot_train_test_matrix(
        self,
        x_train: np.ndarray,
        x_test: np.ndarray,
        save_path: Optional[str | Path] = None,
        n_jobs: int = DEFAULT_N_JOBS,
    ) -> None:
        """Plot kernel matrix for training and testing data.
        Caluculation of kernel values is performed in parallel.

        Args:
            x_train (np.ndarray): array of training samples
            x_test (np.ndarray): array of testing samples
            save_path (Optional[str | Path], optional): save path for the plot. Defaults to None.
            n_jobs (int, optional): number of jobs for parallel computation. Defaults to DEFAULT_N_JOBS.
        """
        train_kernel, _ = self.compute_matrix(x_train, x_train, return_shots_resutls=False, n_jobs=n_jobs)
        test_kernel, _ = self.compute_matrix(x_test, x_train, return_shots_resutls=False, n_jobs=n_jobs)

        fig, axs = plt.subplots(1, 2, figsize=(10, 5))

        train_axs = axs[0]  # type: ignore
        im_train = train_axs.imshow(np.asmatrix(train_kernel), interpolation="nearest", origin="upper", cmap="Blues")
        train_axs.set_title("Training kernel matrix")
        divider_train = make_axes_locatable(train_axs)
        cax_train = divider_train.append_axes("right", size="5%", pad=0.2)
        fig.colorbar(im_train, cax=cax_train)

        test_axs = axs[1]  # type: ignore
        im_test = test_axs.imshow(np.asmatrix(test_kernel), interpolation="nearest", origin="upper", cmap="Reds")
        test_axs.set_title("Testing kernel matrix")
        divider_test = make_axes_locatable(test_axs)
        cax_test = divider_test.append_axes("right", size="5%", pad=0.2)
        fig.colorbar(im_test, cax=cax_test)

        plt.tight_layout()

        if save_path is not None:
            plt.savefig(save_path)

        plt.show()



import types

from qxmt.configs import ExperimentConfig
from qxmt.constants import DEFAULT_N_JOBS
from qxmt.devices.base import BaseDevice
from qxmt.devices.builder import DeviceBuilder
from qxmt.exceptions import InvalidModelNameError
from qxmt.feature_maps.base import BaseFeatureMap
from qxmt.kernels.base import BaseKernel
from qxmt.models.base import BaseMLModel
from qxmt.models.qrigge import QRiggeRegressor
from qxmt.models.qsvc import QSVC
from qxmt.models.qsvr import QSVR
from qxmt.utils import load_object_from_yaml


class ModelBuilder:
    """Builder class for quantum machine learning models.
    This class is responsible for building quantum machine learning models.
    Absorb differences among various platforms, Feature Maps, and Kernels,
    and build models that can be handled as a common interface within the library.

    Examples:
        >>> from qxmt.configs import ExperimentConfig
        >>> from qxmt.models.builder import ModelBuilder
        >>> config = ExperimentConfig(path="configs/my_run.yaml")
        >>> model = ModelBuilder(config).build()
    """

    def __init__(self, config: ExperimentConfig, n_jobs: int = DEFAULT_N_JOBS) -> None:
        """Initialize the model builder.

        Args:
            config (ExperimentConfig): Configuration for the experiment.
            n_jobs (int): number of jobs for parallel computation
        """
        self.config: ExperimentConfig = config
        self.n_jobs: int = n_jobs
        self.device: BaseDevice
        self.feature_map: BaseFeatureMap
        self.kernel: BaseKernel
        self.model: BaseMLModel

    def _set_device(self) -> None:
        """Set quantum device.

        Raises:
            InvalidPlatformError: platform is not implemented.
        """
        device_config = self.config.device
        self.device = DeviceBuilder(config=device_config).build()

    def _set_feature_map(self) -> None:
        """Set quantum feature map."""
        feature_map_config = self.config.feature_map
        if feature_map_config is None:
            return
        else:
            self.feature_map = load_object_from_yaml(
                config=feature_map_config.model_dump(),
                dynamic_params={
                    "n_qubits": self.config.device.n_qubits,
                },
            )

        if not (isinstance(self.feature_map, BaseFeatureMap) or isinstance(self.feature_map, types.FunctionType)):
            raise TypeError("Feature map must be a BaseFeatureMap instance or a function.")

    def _set_kernel(self) -> None:
        """Set quantum kernel."""
        kernel_config = self.config.kernel
        if kernel_config is None:
            return
        else:
            self.kernel = load_object_from_yaml(
                config=kernel_config.model_dump(),
                dynamic_params={
                    "device": self.device,
                    "feature_map": self.feature_map,
                },
            )

        if not isinstance(self.kernel, BaseKernel):
            raise TypeError("Kernel must be a BaseKernel instance.")

    def _set_model(self) -> None:
        """Set quantum model."""
        match self.config.model.name:
            case "qsvc":
                self.model = QSVC(kernel=self.kernel, n_jobs=self.n_jobs, **self.config.model.params)
            case "qsvr":
                self.model = QSVR(kernel=self.kernel, n_jobs=self.n_jobs, **self.config.model.params)
            case "qrigge":
                self.model = QRiggeRegressor(kernel=self.kernel, n_jobs=self.n_jobs, **self.config.model.params)
            case _:
                raise InvalidModelNameError(f'"{self.config.model.name}" is not implemented.')

    def build(self) -> BaseMLModel:
        """
        Build quantum model by following steps:
            1. Set quantum device
            2. Set quantum feature map
            3. Set quantum kernel
            4. Set quantum model
        """
        self._set_device()
        self._set_feature_map()
        self._set_kernel()
        self._set_model()

        return self.model



from pathlib import Path
from typing import Optional, cast

import numpy as np
import pandas as pd


class FileDataLoader:
    """
    This class loads the data and label from the multi format file.
    The loaded data and label are returned as numpy arrays (X and y).

    Input data schema supports two patterns:
    1. data and label are defined in separate files.
    2. data and label are defined in a single file. In this case, the label name must be defined.

    Supported file formats:
    - .npy, .npz, .csv, .tsv

    Examples:
        >>> loader = FileDataLoader(data_path="data.npy", label_path="label.npy")
        >>> X, y = loader.load()
        >>> loader = FileDataLoader(data_path="data.npz")
        >>> X, y = loader.load()
        >>> loader = FileDataLoader(data_path="data.csv", label_path="label.csv")
        >>> X, y = loader.load()
        >>> loader = FileDataLoader(data_path="data.csv", label_name="target")
        >>> X, y = loader.load()
        >>> loader = FileDataLoader(data_path="data.tsv", label_path="label.tsv")
        >>> X, y = loader.load()
        >>> loader = FileDataLoader(data_path="data.tsv", label_name="target")
        >>> X, y = loader.load()
    """

    def __init__(
        self,
        data_path: str | Path,
        label_path: Optional[str | Path] = None,
        label_name: Optional[str] = None,
    ) -> None:
        """Initialize the FileDataLoader.

        Args:
            data_path (str | Path): path to the data file.
            label_path (Optional[str  |  Path], optional): path to the label file. Defaults to None.
            label_name (Optional[str], optional): label name in the dataset. Defaults to None.
        """
        self.data_path = Path(data_path)
        self.label_path = Path(label_path) if label_path is not None else None
        self.label_name = label_name

    def load(self) -> tuple[np.ndarray, np.ndarray]:
        """
        Load the data and label from the file path.
        The file format is determined by the extension of the file path.

        Supported file formats:
        - numpy: .npy, .npz
        - pandas: .csv, .tsv

        Two input patterns exist:
        1. "data_path" and "label_path" are defined.
        "label_name" is not needed, because the label is loaded from the file.
        2. "data_path" and "label_name" are defined.
        "label_path" is not needed, because the label data include in the data file.

        Returns:
            tuple[np.ndarray, np.ndarray]: loaded data and label as numpy arrays.

        Raises:
            ValueError: data and label file extensions do not match
            ValueError: Data or label key is not matched in the npz file.
            ValueError: Label defined in data of "label_path", not need to define "label_name".
            ValueError: "Data and label are expected to be contained in the single file defined in "data_path"
            ValueError: Label name is not found in the dataset.
            ValueError: unsupported file extension
        """
        data_path = Path(self.data_path)
        data_extension = data_path.suffix

        # label path and extension check only executed when label is not None
        if self.label_path is not None:
            label_path = Path(self.label_path)
            label_extension = label_path.suffix
            if data_extension != label_extension:
                raise ValueError("Data and label file extensions do not match.")

        match data_extension:
            case ".npy":
                # input two npy file paths from config
                X = np.load(data_path, allow_pickle=True)
                y = np.load(label_path, allow_pickle=True)
            case ".npz" if self.label_path is None:
                # input one npz file path from config
                data = np.load(data_path, allow_pickle=True)
                X = data.get("X", None)
                y = data.get("y", None)
                if X is None or y is None:
                    raise ValueError(
                        """Data or label key is not matched in the npz file.
                        Data key must be set as 'X' and label key must be set as 'y'."""
                    )
            case ".csv" if self.label_path is not None:
                # input two csv file paths from config
                if self.label_name is not None:
                    raise ValueError('Label defined in data of "label_path", not need to define "label_name".')
                X = pd.read_csv(data_path, sep=",").values
                y = pd.read_csv(label_path, sep=",").values
            case ".csv" if self.label_path is None:
                # input one csv file path from config
                if self.label_path is not None:
                    raise ValueError(
                        """Data and label are expected to be contained in the single file defined in "data_path"
                        , but "label_path" is also defined."""
                    )
                data = pd.read_csv(data_path, sep=",")
                if self.label_name not in data.columns:
                    raise ValueError(f'Label name "{self.label_name}" is not found in the dataset.')
                X = data.drop(columns=[self.label_name]).values
                y = cast(np.ndarray, data[self.label_name].values)
            case ".tsv" if self.label_path is not None:
                # input two tsv file paths from config
                if self.label_name is not None:
                    raise ValueError('Label defined in data of "label_path", not need to define "label_name".')
                X = pd.read_csv(data_path, sep="\t").values
                y = pd.read_csv(label_path, sep="\t").values
            case ".tsv" if self.label_path is None:
                # input one tsv file path from config
                if self.label_path is not None:
                    raise ValueError(
                        """Data and label are expected to be contained in the single file defined in "data_path"
                        , but "label_path" is also defined."""
                    )
                data = pd.read_csv(data_path, sep="\t")
                if self.label_name not in data.columns:
                    raise ValueError(f'Label name "{self.label_name}" is not found in the dataset.')
                X = data.drop(columns=[self.label_name]).values
                y = data[self.label_name].values
            case _:
                raise ValueError(f'Unsupported file extension: "{data_extension}"')

        return X, np.array(y)



from typing import Annotated, Any, Optional

import numpy as np
from pydantic import BaseModel, PlainSerializer, PlainValidator

from qxmt.configs import DatasetConfig


def validate(v: Any) -> np.ndarray:
    if isinstance(v, np.ndarray):
        return v
    else:
        raise TypeError(f"Expected numpy array, got {type(v)}")


def serialize(v: np.ndarray) -> list[list[float]]:
    return v.tolist()


DataArray = Annotated[
    np.ndarray,
    PlainValidator(validate),
    PlainSerializer(serialize),
]


class Dataset(BaseModel):
    # model_config = ConfigDict(frozen=True, extra="forbid")

    X_train: DataArray
    y_train: DataArray
    X_val: Optional[DataArray]
    y_val: Optional[DataArray]
    X_test: DataArray
    y_test: DataArray
    config: DatasetConfig



from collections import Counter
from itertools import product

import numpy as np


def validate_sampling_values(sampling_result: np.ndarray, valid_values: list[int] = [0, 1]) -> None:
    """Validate the sampling resutls of each shots.

    Args:
        sampling_result (np.ndarray): array of sampling results
        valid_values (list[int], optional): valid value of quantum state. Defaults to [0, 1].

    Raises:
        ValueError: invalid values in the sampling results
    """
    if not np.all(np.isin(sampling_result, valid_values)):
        unique_values = np.unique(sampling_result)
        invalid_values = unique_values[~np.isin(unique_values, valid_values)]
        raise ValueError(f"The input array contains values other than 0 and 1. (invalid values: {invalid_values})")


def generate_all_observable_states(n_qubits: int, state_pattern: str = "01") -> list[str]:
    """Generate all possible observable states for the given number of qubits.

    Args:
        n_qubits (int): number of qubits
        state_pattern (str, optional): pattern of the observable state. Defaults to "01".

    Returns:
        list[str]: list of all possible observable states
    """
    return ["".join(bits) for bits in product(state_pattern, repeat=n_qubits)]


def sample_results_to_probs(result: np.ndarray, n_qubits: int, shots: int, state_pattern: str = "01") -> np.ndarray:
    """Convert the sampling results to the probability of each

    Args:
        result (np.ndarray): numpy array of sampling results
        n_qubits (int): number of qubits
        shots (int): number of shots
        state_pattern (str, optional): pattern of the observable state. Defaults to "01".

    Returns:
        np.ndarray: numpy array of the probability of each state
    """
    # validate sampleing results for getting the each state probability
    validate_sampling_values(result)

    # convert the sample results to bit strings
    # ex) shots=3, n_qubits=2, [[0, 0], [1, 1], [0, 0]] => ["00", "11", "00"]
    result = np.array([result]) if result.ndim == 1 else result
    bit_strings = ["".join(map(str, sample)) for sample in result]
    all_states = generate_all_observable_states(n_qubits, state_pattern=state_pattern)

    # count the number of each state
    count_dict = Counter(bit_strings)
    state_counts = [count_dict.get(state, 0) for state in all_states]

    # convert the count to the probability
    probs = np.array(state_counts) / shots  # shots must be over 0

    return probs



from pathlib import Path
from typing import Any, Literal, Optional

import yaml
from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator

from qxmt.constants import PROJECT_ROOT_DIR


class GlobalSettingsConfig(BaseModel):
    model_config = ConfigDict(frozen=True, extra="forbid")
    random_seed: int
    task_type: Literal["classification", "regression"]


class OpenMLConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    name: Optional[str] = None
    id: Optional[int] = None
    return_format: str = "numpy"
    save_path: Optional[Path | str] = None

    def model_post_init(self, __context: dict[str, Any]) -> None:
        if (self.save_path is not None) and (not Path(self.save_path).is_absolute()):
            self.save_path = PROJECT_ROOT_DIR / self.save_path


class FileConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    data_path: Path | str
    label_path: Optional[Path | str]
    label_name: Optional[str]

    def model_post_init(self, __context: dict[str, Any]) -> None:
        if not Path(self.data_path).is_absolute():
            self.data_path = PROJECT_ROOT_DIR / self.data_path

        if (self.label_path is not None) and not Path(self.label_path).is_absolute():
            self.label_path = PROJECT_ROOT_DIR / self.label_path


class GenerateDataConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    generate_method: Literal["linear"]
    params: Optional[dict[str, Any]] = {}


class SplitConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    train_ratio: float = Field(ge=0.0, le=1.0)
    validation_ratio: float = Field(ge=0.0, le=1.0, default=0.0)
    test_ratio: float = Field(ge=0.0, le=1.0)
    shuffle: bool = Field(default=True)

    @model_validator(mode="after")
    def check_ratio(self) -> "SplitConfig":
        ratios = [self.train_ratio, self.validation_ratio, self.test_ratio]
        if sum(ratios) != 1:
            raise ValueError("The sum of the ratios must be 1.")

        return self


class DatasetConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    openml: Optional[OpenMLConfig] = None  # only need when use openml dataset
    file: Optional[FileConfig] = None  # only need when use file dataset
    generate: Optional[GenerateDataConfig] = None  # only need when use generated dataset
    random_seed: int
    split: SplitConfig
    features: Optional[list[str]] = None
    raw_preprocess_logic: Optional[list[dict[str, Any]] | dict[str, Any]] = None
    transform_logic: Optional[list[dict[str, Any]] | dict[str, Any]] = None


class DeviceConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    platform: str
    name: str
    n_qubits: int
    shots: Optional[int] = None
    random_seed: Optional[int] = None
    save_shots_results: bool = False

    @field_validator("shots")
    def check_shots(cls, value: int) -> int:
        if (value is not None) and (value < 1):
            raise ValueError("shots must be greater than or equal to 1")
        return value

    @model_validator(mode="after")
    def check_save_shots(self) -> "DeviceConfig":
        if (self.shots is None) and (self.save_shots_results):
            raise ValueError('The "shots" must be set to save the shot results.')
        return self


class FeatureMapConfig(BaseModel):
    model_config = ConfigDict(frozen=True, extra="forbid")

    module_name: str
    implement_name: str
    params: Optional[dict[str, Any]] = None


class KernelConfig(BaseModel):
    model_config = ConfigDict(frozen=True, extra="forbid")

    module_name: str
    implement_name: str
    params: Optional[dict[str, Any]] = None


class ModelConfig(BaseModel):
    model_config = ConfigDict(frozen=True, extra="forbid")

    name: str
    params: dict[str, Any]
    feature_map: Optional[FeatureMapConfig] = None
    kernel: Optional[KernelConfig] = None


class EvaluationConfig(BaseModel):
    model_config = ConfigDict(frozen=True, extra="forbid")

    default_metrics: list[str]
    custom_metrics: Optional[list[dict[str, Any]]] = None


class ExperimentConfig(BaseModel):
    model_config = ConfigDict(frozen=True, extra="forbid")

    path: Path | str = ""
    description: str = ""
    global_settings: GlobalSettingsConfig
    dataset: DatasetConfig
    device: DeviceConfig
    feature_map: Optional[FeatureMapConfig] = None
    kernel: Optional[KernelConfig] = None
    model: ModelConfig
    evaluation: EvaluationConfig

    def __init__(self, **data: Any) -> None:
        """Initialize the experiment configuration.

        Case 1:
            Load the configuration from a file path.
            This case the data is a dictionary with a single key "path".
        Case 2:
            Load the configuration from a dictionary.
            This case the data is a dictionary with the configuration data.
        """
        if list(data.keys()) == ["path"]:
            config = self.load_from_path(data.get("path", ""))
            data.update(config)
        super().__init__(**data)

    def load_from_path(self, path: str) -> dict[str, Any]:
        with open(path, "r") as file:
            config = yaml.safe_load(file)

        if config is None:
            raise ValueError(f'The configuration file is empty. (path="{path}")')

        return {
            "description": config.get("description"),
            "global_settings": config.get("global_settings"),
            "dataset": config.get("dataset"),
            "device": config.get("device"),
            "feature_map": config.get("feature_map"),
            "kernel": config.get("kernel"),
            "model": config.get("model"),
            "evaluation": config.get("evaluation"),
        }



from typing import Optional

import numpy as np


def generate_linear_separable_data(
    n_samples: int = 100,
    n_features: int = 2,
    n_classes: int = 2,
    noise: float = 0.1,
    scale: float = 1.0,
    random_seed: Optional[int] = None,
) -> tuple[np.ndarray, np.ndarray]:
    """Generate random linear separable data for multi-class classification.

    Args:
        n_samples (int, optional): sample size. Defaults to 100.
        n_features (int, optional): dimension of feature. Defaults to 2.
        n_classes (int, optional): number of classes. Defaults to 2.
        noise (float, optional): noise level. Defaults to 0.1.
        scale (float, optional): scale of data. Defaults to 1.0.
        random_seed (int, optional): random seed. Defaults to None.

    Returns:
        tuple[np.ndarray, np.ndarray]: generated data and label
    """
    # set the random seed for reproducibility
    if random_seed is not None:
        np.random.seed(random_seed)

    X = scale * np.random.randn(n_samples, n_features)
    w = scale * np.random.randn(n_classes, n_features)

    # generate scores by multiplying feature matrix and weight matrix
    scores = X @ w.T + noise * np.random.randn(n_samples, n_classes)

    # assign the class with the highest score as the label
    y = np.argmax(scores, axis=1)

    return X, y


def generate_linear_regression_data(
    n_samples: int = 100,
    n_features: int = 2,
    noise: float = 0.1,
    scale: float = 1.0,
    random_seed: Optional[int] = None,
) -> tuple[np.ndarray, np.ndarray]:
    """Generate random data for linear regression.

    Args:
        n_samples (int, optional): Number of samples. Defaults to 100.
        n_features (int, optional): Number of features. Defaults to 2.
        noise (float, optional): Noise level to add to the target. Defaults to 0.1.
        scale (float, optional): Scale of the data. Defaults to 1.0.
        random_seed (int, optional): Random seed for reproducibility. Defaults to None.

    Returns:
        Tuple[np.ndarray, np.ndarray]: Generated features and target values
    """
    # set the random seed for reproducibility
    if random_seed is not None:
        np.random.seed(random_seed)

    # generate random feature data
    X = scale * np.random.randn(n_samples, n_features)

    # generate random weights for the linear relationship
    true_weights = scale * np.random.randn(n_features)

    # generate target variable with a linear relationship to the features
    y = X @ true_weights + noise * np.random.randn(n_samples)

    return X, y



import copy
from pathlib import Path
from typing import Any, Optional, cast

import dill
import numpy as np
from sklearn.kernel_ridge import KernelRidge
from sklearn.model_selection import cross_val_score

from qxmt.constants import DEFAULT_N_JOBS
from qxmt.kernels.base import BaseKernel
from qxmt.models.base import BaseKernelModel
from qxmt.models.hyperparameter_search.search import HyperParameterSearch


class QRiggeRegressor(BaseKernelModel):
    """Quantum Rigge Regressor model.

    Detail of Kernel Rigge Regressor model is refered to the sklearn.kernel_ridge.KernelRidge class.
    Reference: https://scikit-learn.org/stable/modules/generated/sklearn.kernel_ridge.KernelRidge.html

    Args:
        BaseKernelModel (_type_): kernel instance of BaseKernel class
    """

    def __init__(self, kernel: BaseKernel, n_jobs: int = DEFAULT_N_JOBS, alpha: float = 1.0, **kwargs: Any) -> None:
        """Initialize the QRiggeRegressor model.

        Args:
            kernel (BaseKernel): kernel instance of BaseKernel class
            n_jobs (int): number of jobs for parallel computation
            alpha (float): Regularization strength; must be a positive float
        """
        super().__init__(kernel, n_jobs)
        self.fit_X: Optional[np.ndarray] = None
        self.model = KernelRidge(alpha=alpha, kernel="precomputed", **kwargs)

    def __getattr__(self, name: str) -> Any:
        # if the attribute is in the model, return it
        if hasattr(self.model, name):
            return getattr(self.model, name)
        # if the attribute is not in the model, raise AttributeError
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    def cross_val_score(
        self,
        X: np.ndarray,
        y: np.ndarray,
        **kwargs: Any,
    ) -> np.ndarray:
        """Cross validation score of the Quantum Kernel Ridge Regressor model.
        Default to use the R^2 score.

        Args:
            X (np.ndarray): numpy array of input data
            y (np.ndarray): numpy array of target data
            **kwargs (dict): additional arguments

        Returns:
            np.ndarray: array of scores of the estimator for each run of the cross validation
        """
        kernel_X, _ = self.kernel.compute_matrix(X, X, return_shots_resutls=False, n_jobs=self.n_jobs)
        scores = cross_val_score(estimator=self.model, X=kernel_X, y=y, n_jobs=self.n_jobs, **kwargs)
        return scores

    def hyperparameter_search(
        self,
        X: np.ndarray,
        y: np.ndarray,
        search_type: str,
        search_space: dict[str, list[Any]],
        search_args: dict[str, Any],
        objective: Optional[Any] = None,
        refit: bool = True,
    ) -> dict[str, Any]:
        """Search the best hyperparameters for the Quantum Kernel Ridge Regressor model.

        Args:
            X (np.ndarray): dataset for search
            y (np.ndarray): target values for search
            search_type (str): search type for hyperparameter search (grid, random, tpe)
            search_space (dict[str, list[Any]]): search space for hyperparameter search
            search_args (dict[str, Any]): search arguments for hyperparameter search
            objective (Optional[Callable], optional): objective function for search. Defaults to None.
            refit (bool, optional): refit the model with best hyperparameters. Defaults to True.

        Raises:
            ValueError: Not supported search type

        Returns:
            dict[str, Any]: best hyperparameters
        """
        search_model = copy.deepcopy(self.model)
        X_kernel, _ = self.kernel.compute_matrix(X, X, return_shots_resutls=False, n_jobs=self.n_jobs)

        if "scoring" not in search_args.keys():
            search_args["scoring"] = "r2"

        searcher = HyperParameterSearch(
            X=X_kernel,
            y=y,
            model=search_model,
            search_type=search_type,
            search_space=search_space,
            search_args=search_args,
            objective=objective,
        )
        best_params = searcher.search()

        if refit:
            self.model.set_params(**best_params)
            self.model.fit(X_kernel, y)

        return best_params

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray,
        save_shots_path: Optional[Path | str] = None,
    ) -> None:
        """Fit the model with given input data and target data.

        Args:
            X (np.ndarray): numpy array of input data
            y (np.ndarray): numpy array of target data
            save_shots_path (Optional[Path | str], optional): save path for the shot results. Defaults to None.
        """
        self.fit_X = X
        if save_shots_path is not None:
            kernel_train_X, shots_matrix = self.kernel.compute_matrix(
                self.fit_X, self.fit_X, return_shots_resutls=True, n_jobs=self.n_jobs, bar_label="Train"
            )
            if shots_matrix is not None:
                self.kernel.save_shots_results(shots_matrix, save_shots_path)
        else:
            kernel_train_X, _ = self.kernel.compute_matrix(
                self.fit_X, self.fit_X, return_shots_resutls=False, n_jobs=self.n_jobs, bar_label="Train"
            )
        self.model.fit(kernel_train_X, y)

    def predict(self, X: np.ndarray) -> np.ndarray:
        """Predict target values of input data.

        Args:
            X (np.ndarray): numpy array of input data

        Returns:
            np.ndarray: numpy array of predicted target values
        """
        if self.fit_X is None:
            raise ValueError("The model is not trained yet.")
        else:
            kernel_pred_X, _ = self.kernel.compute_matrix(
                X, self.fit_X, return_shots_resutls=False, n_jobs=self.n_jobs, bar_label="Test"
            )
        return self.model.predict(kernel_pred_X)

    def score(self, X: np.ndarray, y: np.ndarray, sample_weight: Optional[np.ndarray] = None) -> float:
        """Return the coefficient of determination R^2 of the prediction.

        Args:
            X (np.ndarray): numpy array of input data
            y (np.ndarray): numpy array of target data

        Returns:
            float: R^2 of the prediction
        """
        if self.fit_X is None:
            raise ValueError("The model is not trained yet.")
        else:
            kernel_pred_X, _ = self.kernel.compute_matrix(X, self.fit_X, return_shots_resutls=False, n_jobs=self.n_jobs)
        return cast(float, self.model.score(kernel_pred_X, y, sample_weight=sample_weight))

    def save(self, path: str | Path) -> None:
        """Save the model to the given path.

        Args:
            path (str | Path): path to save the model
        """
        dill.dump(self.model, open(path, "wb"))

    def load(self, path: str | Path) -> "QRiggeRegressor":
        """Load the trained model from the given path.

        Args:
            path (str | Path): path to load the model

        Returns:
            QRiggeRegressor: loaded QRiggeRegressor model
        """
        return dill.load(open(path, "rb"))

    def get_params(self) -> dict:
        """Get the parameters of the model."""
        return self.model.get_params()

    def set_params(self, params: dict) -> None:
        """Set the parameters of the model."""
        self.model.set_params(**params)



from qxmt.utils.git import (
    get_commit_id,
    get_git_add_code,
    get_git_diff,
    get_git_rm_code,
    is_git_available,
)
from qxmt.utils.yaml import (
    load_object_from_yaml,
    load_yaml_config,
    save_experiment_config_to_yaml,
)

__all__ = [
    "get_commit_id",
    "get_git_diff",
    "get_git_add_code",
    "get_git_rm_code",
    "is_git_available",
    "load_yaml_config",
    "load_object_from_yaml",
    "save_experiment_config_to_yaml",
]



from qxmt.models.base import BaseKernelModel, BaseMLModel
from qxmt.models.builder import ModelBuilder
from qxmt.models.qrigge import QRiggeRegressor
from qxmt.models.qsvc import QSVC
from qxmt.models.qsvr import QSVR

__all__ = ["BaseMLModel", "BaseKernelModel", "ModelBuilder", "QRiggeRegressor", "QSVC", "QSVR"]
